/*  Programmer:	Piotr Borek
 *  E-mail: piotrborek@op.pl
 *  Copyright 2009 Piotr Borek
 *
 *  Distributed under the terms of the GPL (GNU Public License)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jarre.app;

import static jarre.i18n.I18n._;
import jarre.Jarre;
import jarre.app.widgets.Controls;
import jarre.app.widgets.PianoAbstract;
import jarre.app.widgets.PianoKeyboard;
import jarre.app.widgets.PianoTape;
import jarre.app.widgets.PlayerWidget;
import jarre.app.widgets.Controls.SequenceChange;
import jarre.midi.MidiSystem;
import jarre.midi.MySequence;
import jarre.midi.PianoReceiver;
import jarre.midi.event.NoteOffEvent;
import jarre.midi.event.NoteOnEvent;
import jarre.midi.sequencer.PianoSequencer;
import jarre.midi.sequencer.SeqDrawEvent;
import jarre.midi.sequencer.SeqNoteEvent;
import jarre.midi.sequencer.SeqTrackEvent;
import jarre.widgets.PlayerButtons.Action;
import jarre.widgets.PlayerButtons.ActionListener;
import jarre.widgets.common.App;
import jarre.widgets.common.IProgressBar;
import jarre.widgets.utils.GridLayoutFactory;
import jarre.widgets.windows.AboutWindow;
import jarre.widgets.windows.AppWindow;

import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.CTabFolder;
import org.eclipse.swt.custom.CTabItem;
import org.eclipse.swt.custom.SashForm;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;
import org.eclipse.swt.widgets.MessageBox;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.ToolBar;
import org.eclipse.swt.widgets.ToolItem;

public class MainWindow extends AppWindow {
    private CTabFolder folder;
    private Controls controls;
    private ToolItem showHideItem;
    private PianoSequencer pianoSequencer;
    private PianoTape pianoTape;
    private PianoKeyboard keyboard;
    private PianoReceiver pianoReceiver;
    private MySequence sequence;
    private boolean shrinkKeyboard = false;
    public MainWindow(Shell shell, String title) {
	super(shell, title, App.CREATE_MENU_BAR);
	Listener listener = new Listener() {
	    @Override public void handleEvent(Event event) {
		switch (event.keyCode) {
		    case 'f':
		    case 'F':
			toggleFullscreen();
			break;
		    case 'p':
		    case 'P':
			if (pianoSequencer.isPlaying()) execAction(Action.PAUSE);
			else execAction(Action.PLAY);
			break;
		    case 'm':
		    case 'M':
			toggleFollowMode(null);
			break;
		    case '=':
		    case '+':
			increaseSpeed();
			break;
		    case '-':
		    case '_':
			decreaseSpeed();
			break;
		    case '[':
		    case '{':
			zoomOut();
			break;
		    case ']':
		    case '}':
			zoomIn();
			break;
		    case '1':
		    case '!':
			zoomOriginal();
			break;
		    case 's':
		    case 'S':
			toggleShrinkButton(null);
			break;
		    case 'i':
		    case 'I':
			toggleIgnoreProgram(null);
			break;
		    case 'b':
		    case 'B':
			execAction(Action.START);
			break;
		    case 'h':
		    case 'H':
			toggleControls();
			break;
		    case 't':
		    case 'T':
			toggleMetronome(null);
			break;
		    case 'o':
		    case 'O':
			controls.openFileDialog();
			break;
		    case SWT.ARROW_DOWN:
			execAction(Action.BACKWARD);
			break;
		    case SWT.ARROW_UP:
			execAction(Action.FORWARD);
			break;
		    case SWT.ARROW_LEFT:
			if (sequence == null) break;
			pianoSequencer.goBackward(sequence.getResolution());
			break;
		    case SWT.ARROW_RIGHT:
			if (sequence == null) break;
			pianoSequencer.goForward(sequence.getResolution());
			break;
		}
	    }
	};
	setKeyDownListeners(getShell().getChildren(), listener);
    }
    private void setKeyDownListeners(Control[] array, Listener listener) {
	if (array.length == 0) return;
	for (Control control: array) {
	    if (control instanceof Composite) {
		Composite composite = (Composite)control;
		composite.addListener(SWT.KeyDown, listener);
		setKeyDownListeners(composite.getChildren(), listener);
	    } else control.addListener(SWT.KeyDown, listener);
	}
    }
    public PianoKeyboard getPianoKeyboard() {
	return keyboard;
    }
    @Override public void createMenuBar(Menu menu) {
	MenuItem item = new MenuItem (menu, SWT.CASCADE);
	item.setText (_("Options"));
	Menu subMenu = new Menu(getShell(), SWT.DROP_DOWN);
	item.setMenu(subMenu);
	item = new MenuItem(subMenu, SWT.PUSH);
	item.setText(_("Fullscreen"));
	item.addListener(SWT.Selection, new Listener() {
	    @Override public void handleEvent(Event event) {
		toggleFullscreen();
	    }
	});
	new MenuItem(subMenu, SWT.SEPARATOR);
	item = new MenuItem(subMenu, SWT.PUSH);
	item.setText(_("Configure MIDI"));
	item.addListener(SWT.Selection, new Listener() {
	    @Override public void handleEvent(Event event) {
		ConfigWindow config = new ConfigWindow(getShell(), _("MIDI configuration"));
		if (config.open() == SWT.OK) MidiSystem.initialize();
	    }
	});
	item = new MenuItem(subMenu, SWT.PUSH);
	item.setText(_("Reset the MIDI driver"));
	item.addListener(SWT.Selection, new Listener() {
	    @Override public void handleEvent(Event event) {
		pianoSequencer.pause();
		controls.getPlayerWidget().setPlay(false);
		MidiSystem.allSoundOff();
		MidiSystem.initialize();
		MidiSystem.allSoundOff();
		MessageBox box = new MessageBox(getShell(), SWT.ICON_INFORMATION | SWT.OK);
		box.setMessage(_("The MIDI driver was successfuly reseted."));
		box.open();
	    }
	});
	new MenuItem(subMenu, SWT.SEPARATOR);
	item = new MenuItem(subMenu, SWT.PUSH);
	item.setText(_("About"));
	item.addListener (SWT.Selection, new Listener () {
	    @Override public void handleEvent (Event e) {
		AboutWindow about = new AboutWindow();
		about.open();
	    }
	});
    }
    public void pause() {
	execAction(Action.PAUSE);
    }
    @Override protected void createControl(Composite parent) {
	GridLayoutFactory.swtDefaults().columns(2).applyTo(parent);
	final SashForm sashForm = new SashForm(parent, SWT.HORIZONTAL);
	sashForm.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
	folder = new CTabFolder(sashForm, SWT.BORDER | SWT.FLAT);
	folder.setSimple(false);
	folder.setFocus();
	controls = new Controls(folder, SWT.NONE);
	controls.addSequenceListener(new SequenceChange() {
	    @Override public void hanldeEvent(MySequence sequence) {
		setSequence(sequence);
		controls.getPlayerWidget().setCurrentTime(0);
	    }
	});
	controls.addActionListener(new ActionListener() {
	    @Override public void handleEvent(Action action) {
		execAction(action);
	    }
	});
	controls.addMetronomeListener(new Listener() {
	    @Override public void handleEvent(Event event) {
		toggleMetronome((Button)event.widget);
	    }
	});
	controls.addFollowListener(new Listener() {
	    @Override public void handleEvent(Event event) {
		toggleFollowMode((Button)event.widget);
	    }
	});
	controls.addShrinkListener(new Listener() {
	    @Override public void handleEvent(Event event) {
		toggleShrinkButton((Button)event.widget);
	    }
	});
	controls.addIgnoreListener(new Listener() {
	    @Override public void handleEvent(Event event) {
		toggleIgnoreProgram((Button)event.widget);
	    }
	});
	controls.addTrackChangeListener(new Listener() {
	    @Override public void handleEvent(Event event) {
		if (pianoSequencer.getFollowMode()) pianoSequencer.noteChangeSignal();
		pianoTape.redrawAndUpdate();
		keyboard.redrawAndUpdate();
	    }
	});
	controls.addProgressBarListener(new Listener() {
	    @Override public void handleEvent(Event event) {
		IProgressBar progressBar = (IProgressBar)event.widget;
		pianoSequencer.goToPosition(progressBar.getSelection());
	    }
	});
	CTabItem item = new CTabItem(folder, SWT.NONE);
	item.setText(_("Controls"));
	item.setControl(controls);
	Composite composite = new Composite(sashForm, SWT.NONE);
	GridLayoutFactory.fillDefaults().applyTo(composite);
	pianoTape = new PianoTape(composite, SWT.NONE);
	pianoTape.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
	keyboard = new PianoKeyboard(composite, SWT.NONE);
	keyboard.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false));
	pianoSequencer = new PianoSequencer(keyboard);
	pianoSequencer.addNoteEvent(new SeqNoteEvent() {
	    private Runnable keyboardRedraw = new Runnable() {
		@Override public void run() {
		    keyboard.redrawAndUpdate();
		}
	    };
	    @Override public void allNotesOffEvent() {
		if (! keyboard.isDisposed())
		    keyboard.setAllNotesOff();
	    }
	    @Override public void noteOffEvent(NoteOffEvent noteOff) {
		keyboard.setNoteOff(noteOff);
		if (!pianoSequencer.isPlaying() || pianoSequencer.isWaiting())
		    getDisplay().asyncExec(keyboardRedraw);
	    }
	    @Override public void noteOnEvent(NoteOnEvent noteOn) {
		keyboard.setNoteOn(noteOn);
		if (!pianoSequencer.isPlaying() || pianoSequencer.isWaiting())
		    getDisplay().asyncExec(keyboardRedraw);
	    }
	});
	pianoSequencer.addTrackEvent(new SeqTrackEvent() {
	    @Override public void endOfTracksEvent() {
		execAction(Action.PAUSE);
	    }
	    @Override public void sequenceChanged(MySequence sequence) {
	    }
	});
	pianoSequencer.addDrawEvent(new SeqDrawEvent() {
	    private final PlayerWidget playerWidget = controls.getPlayerWidget();
	    private int time = 0;
	    @Override public void drawSequence(int currentPos) {
		updateTime(currentPos);
		keyboard.checkAndUpdate();
	    }
	    @Override public void moveSequence(int currentPos) {
		updateTime(currentPos);
		keyboard.checkAndUpdate();
	    }
	    private void updateTime(int currentPos) {
		final int time = pianoSequencer.getTime() / 1000;
		if (this.time != time) {
		    this.time = time;
		    playerWidget.setCurrentTime(time);
		    playerWidget.update();
		}
	    }
	});
	pianoReceiver = new PianoReceiver(keyboard, pianoSequencer);
	MidiSystem.getTransmitter().setReceiver(pianoReceiver);
	pianoSequencer.setReceiver(MidiSystem.getReceiver());
	pianoTape.setSequencer(pianoSequencer);
	createToolBar(parent);
	getShell().addListener(SWT.Resize, new Listener() {
	    int minWidth = Integer.MAX_VALUE;
	    @Override public void handleEvent(Event event) {
		int width = folder.computeSize(SWT.DEFAULT, SWT.DEFAULT).x + 15;
		if (minWidth > width) minWidth = width;
		sashForm.setWeights(new int[] { minWidth, getShell().getSize().x - minWidth} );
	    }
	});
    }
    protected void createToolBar(Composite parent) {
	ToolBar toolBar = new ToolBar(parent, SWT.VERTICAL);
	toolBar.setLayoutData(new GridData(SWT.TOP, SWT.FILL, false, true));
	ToolItem toolItem = new ToolItem(toolBar, SWT.PUSH);
	toolItem.setText("<");
	toolItem.setToolTipText(_("Hide controls.\nShortcut: h"));
	toolItem.addListener(SWT.Selection, new Listener() {
	    @Override public void handleEvent(Event event) {
		toggleControls();
	    }
	});
	showHideItem = toolItem;
	new ToolItem(toolBar, SWT.SEPARATOR);
	toolItem = new ToolItem(toolBar, SWT.PUSH);
	toolItem.setImage(new Image(getDisplay(), "data/view-fullscreen.png"));
	toolItem.addListener(SWT.Selection, new Listener() {
	    @Override public void handleEvent(Event event) {
		toggleFullscreen();
	    }
	});
	toolItem.setToolTipText(_("Shortcut: f"));
	new ToolItem(toolBar, SWT.SEPARATOR);
	toolItem = new ToolItem(toolBar, SWT.PUSH);
	toolItem.setImage(new Image(getDisplay(), "data/zoom-original.png"));
	toolItem.addListener(SWT.Selection, new Listener() {
	    @Override public void handleEvent(Event event) {
		zoomOriginal();
	    }
	});
	toolItem.setToolTipText(_("Shortcut: 1"));
	toolItem = new ToolItem(toolBar, SWT.PUSH);
	toolItem.setImage(new Image(getDisplay(), "data/zoom-in.png"));
	toolItem.addListener(SWT.Selection, new Listener() {
	    @Override public void handleEvent(Event event) {
		zoomIn();
	    }
	});
	toolItem.setToolTipText(_("Shortcut: ]"));
	toolItem = new ToolItem(toolBar, SWT.PUSH);
	toolItem.setImage(new Image(getDisplay(), "data/zoom-out.png"));
	toolItem.addListener(SWT.Selection, new Listener() {
	    @Override public void handleEvent(Event event) {
		zoomOut();
	    }
	});
	toolItem.setToolTipText(_("Shortcut: ["));
	new ToolItem(toolBar, SWT.SEPARATOR);
	controls.getPlayerWidget().createButtons(toolBar, ToolItem.class);
	new ToolItem(toolBar, SWT.SEPARATOR);
	toolItem = new ToolItem(toolBar, SWT.PUSH);
	toolItem.setImage(new Image(getDisplay(), "data/up.png"));
	toolItem.addListener(SWT.Selection, new Listener() {
	    @Override public void handleEvent(Event event) {
		increaseSpeed();
	    }
	});
	toolItem.setToolTipText(_("Shortcut: +"));
	toolItem = new ToolItem(toolBar, SWT.PUSH);
	toolItem.setImage(new Image(getDisplay(), "data/down.png"));
	toolItem.addListener(SWT.Selection, new Listener() {
	    @Override public void handleEvent(Event event) {
		decreaseSpeed();
	    }
	});
	toolItem.setToolTipText(_("Shortcut: -"));
    }
    @Override protected void closeShell(Event event) {
	pianoSequencer.pause();
	MidiSystem.closeMidi();
    }
    private void toggleFullscreen() {
	setFullscreen(! isFullscreen());
    }
    private void execAction(Action action) {
	switch (action) {
	    case PLAY:
		if (sequence != null) {
		    pianoSequencer.play();
		    controls.getPlayerWidget().setPlay(true);
		} else controls.getPlayerWidget().setPlay(false);
		break;
	    case PAUSE:
		pianoSequencer.pause();
		controls.getPlayerWidget().setPlay(false);
		break;
	    case START:
		pianoSequencer.goToStart();
		break;
	    case BACKWARD:
		pianoSequencer.goBackward();
		break;
	    case FORWARD:
		pianoSequencer.goForward();
		break;
	}
    }
    private void zoomIn() {
	pianoTape.setTapeScale(pianoTape.getTapeScale() * 2);
    }
    private void zoomOut() {
	pianoTape.setTapeScale(pianoTape.getTapeScale() / 2);
    }
    private void zoomOriginal() {
	pianoTape.setTapeScale(100);
    }
    private void increaseSpeed() {
	int speed = pianoSequencer.getSpeed();
	if (speed >= 50) speed = speed + 10;
	else if (speed >= 10) speed = speed + 5;
	else speed = speed + 1;
	pianoSequencer.setSpeed(speed);
	pianoTape.setSpeed(speed);
    }
    private void decreaseSpeed() {
	int speed = pianoSequencer.getSpeed();
	if (speed > 50) speed = speed - 10;
	else if (speed > 10) speed = speed - 5;
	else speed = speed - 1;
	pianoSequencer.setSpeed(speed);
	pianoTape.setSpeed(speed);
    }
    private void toggleMetronome(Button button) {
	if (button == null) {
	    button = controls.getMetronomeButton();
	    button.setSelection(! button.getSelection());
	}
	pianoSequencer.setMetronome(button.getSelection());
    }
    private void toggleFollowMode(Button button) {
	if (button == null) {
	    button = controls.getFollowButton();
	    button.setSelection(! button.getSelection());
	}
	pianoSequencer.setFollowMode(button.getSelection());
    }
    private void toggleShrinkButton(Button button) {
	if (button == null) {
	    button = controls.getShrinkButton();
	    button.setSelection(! button.getSelection());
	}
	if (button.getSelection()) shrinkKeyboard = true;
	else shrinkKeyboard = false;
	if (sequence != null) setSequence(sequence);
    }
    private void toggleIgnoreProgram(Button button) {
	if (button == null) {
	    button = controls.getIgnoreButton();
	    button.setSelection(! button.getSelection());
	}
	pianoSequencer.setIgnoreProgramChange(button.getSelection());
    }
    private void toggleControls() {
	if (folder.getVisible()) {
	    folder.setVisible(false);
	    showHideItem.setText(">");
	    showHideItem.setToolTipText(_("Show controls.\nShortcut: h"));
	} else {
	    folder.setVisible(true);
	    showHideItem.setText("<");
	    showHideItem.setToolTipText(_("Hide controls.\nShortcut: h"));
	}
	Jarre.window.getShell().layout(true, true);
    }
    private void setSequence(MySequence sequence) {
	if (this.sequence != sequence) {
	    this.sequence = sequence;
	    pianoSequencer.setSequence(sequence);
	    keyboard.setAllNotesOff();
	}
	if (shrinkKeyboard) PianoAbstract.shrinkKeyboard(sequence);
	else PianoAbstract.shrinkKeyboard(null);
	getShell().layout(true, true);
    }
}
