/*
 *	Programmer:	Piotr Borek
 *	E-mail:		piotrborek@op.pl
 *	Copyright 2010  Piotr Borek
 *
 *	Distributed under the terms of the GPL (GNU Public License)
 *
 *   Jarre is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jarre.app.widgets;

import jarre.Config;
import jarre.context.AbstractContext;
import jarre.context.IContext;
import jarre.midi.MidiSystem;
import jarre.midi.MySequence;
import jarre.midi.sequencer.MidiSequencer;
import jarre.midi.sequencer.SeqDrawEvent;
import jarre.widgets.PlayerButtons.Action;
import jarre.widgets.PlayerButtons.ActionListener;
import jarre.widgets.common.IProgressBar;
import jarre.widgets.utils.GridLayoutFactory;
import jarre.widgets.views.DirectoryView;

import java.io.File;

import javax.sound.midi.InvalidMidiDataException;

import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Dialog;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Shell;

public class DirectoryDialog extends Dialog {
    public DirectoryDialog(Shell parent, int style) {
	super(parent, style);
	this.style = style;
    }
    public void setFileFilter(String[] regex) {
	this.regex = regex;
	if (view != null)
	    view.setFileFilter(regex);
    }
    public String open() {
	Shell parent = getParent();
        shell = new Shell(parent, SWT.SHELL_TRIM | SWT.APPLICATION_MODAL);
        shell.setText(getText());
        createControl(shell);
        shell.open();
        Display display = parent.getDisplay();
        while (! shell.isDisposed())
            if (! display.readAndDispatch()) display.sleep();
        Config.setPropertyObject("directory-dialog.directories", view.getUserDirectories());
        Config.setPropertyObject("directory-dialog.recently-used", view.getRecentlyUsed());
        Config.xmlConfig.setProperty("directory-dialog.current", view.getDirectory());
        midiSequencer.pause();
        midiSequencer.getReceiver().close();
        shell.dispose();
        if (view.getSelection() != null) return view.getSelection().getAbsolutePath();
        else return null;
    }
    public File getSelection() {
	return view.getSelection();
    }
    protected void createControl(final Shell shell) {
	shell.setLayout(new GridLayout(2, false));
	String dir = Config.xmlConfig.getString("directory-dialog.current", System.getProperty("user.home"));
	Listener closeShellListener = new Listener() {
	    @Override public void handleEvent(Event event) {
		shell.close();
	    }
	};
	view = new DirectoryView(shell, style, dir);
	view.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
	view.getCancelButton().addListener(SWT.Selection, closeShellListener);
	view.getExtraButton().addListener(SWT.Selection, closeShellListener);
	view.addSelectionListener(new Listener() {
	    @Override public void handleEvent(Event event) {
		File file = view.getSelection();
		if (file == null) {
		    if (midiSequencer.getSequence() == null) playerWidget.setEnabled(false);
		    else playerWidget.setEnabled(true);
		    return;
		}
		songLabel.setText(file.getName());
		songLabel.update();
		playerWidget.setEnabled(true);
		try {
		    MySequence sequence = MySequence.create(file);
		    playerContext.enter();
		    midiSequencer.setSequence(sequence);
		    playerWidget.setSongTime((int)(sequence.getMicrosecondLength() / 1000000));
		    playerWidget.setCurrentTime(0);
		    playerContext.exit();
		} catch (InvalidMidiDataException e) {
		    playerWidget.setEnabled(false);
		}
	    }	    
	});
	view.showLocations(true);
	view.showDirPath(true);
	view.setFileFilter(regex);
	view.setUserDirectories((String[])Config.getPropertyObject("directory-dialog.directories"));
	view.setRecentlyUsed((String[])Config.getPropertyObject("directory-dialog.recently-used"));
	Composite composite = new Composite(shell, SWT.NONE);
	GridLayoutFactory.swtDefaults().applyTo(composite);
	playerWidget = new PlayerWidget(composite, SWT.NONE);
	GridData data = new GridData(SWT.FILL, SWT.CENTER, false, false);
	data.widthHint = 160;
	playerWidget.setLayoutData(data);
	playerWidget.setEnabled(false);
	playerWidget.addActionListener(new ActionListener() {
	    @Override public void handleEvent(Action action) {
		switch (action) {
		    case PLAY:
			midiSequencer.play();
			break;
		    case PAUSE:
			midiSequencer.pause();
			break;
		    case START:
			midiSequencer.goToStart();
			break;
		    case BACKWARD:
			midiSequencer.goBackward();
			break;
		    case FORWARD:
			midiSequencer.goForward();
			break;
		}
	    }
	});
	playerWidget.addProgressBarListener(new Listener() {
	    @Override public void handleEvent(Event event) {
		IProgressBar progressBar = (IProgressBar)event.widget;
		midiSequencer.goToPosition(progressBar.getSelection());
	    }
	});
	songLabel = new Label(composite, SWT.NONE);
	songLabel.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false));
	midiSequencer = new MidiSequencer();
	midiSequencer.setReceiver(MidiSystem.getReceiver());
	midiSequencer.addDrawEvent(new SeqDrawEvent() {
	    private int time = 0;
	    @Override public void drawSequence(int currentPos) {
		updateTime(currentPos);
	    }
	    @Override public void moveSequence(int currentPos) {
		updateTime(currentPos);
	    }
	    private void updateTime(int currentPos) {
		final int time = midiSequencer.getTime() / 1000;
		if (this.time != time) {
		    this.time = time;
		    playerWidget.setCurrentTime(time);
		    playerWidget.update();
		}
	    }
	});
    }
    private String[] regex;
    private DirectoryView view;
    private PlayerWidget playerWidget;
    private Label songLabel;
    private Shell shell;
    private int style;
    private MidiSequencer midiSequencer;
    private IContext<Boolean> playerContext = new AbstractContext<Boolean>() {
	@Override public void enter() {
	    if (midiSequencer.isPlaying()) {
		midiSequencer.pause();
		pushContext(true);
	    } else {
		pushContext(false);
	    }
	}
	@Override public void exit() {
	    Boolean p = popContext();
	    if (p == null) return;
	    if (p) midiSequencer.play();   
	}
    };
}
