/*  Programmer:	Piotr Borek
 *  E-mail: piotrborek@op.pl
 *  Copyright 2009 Piotr Borek
 *
 *  Distributed under the terms of the GPL (GNU Public License)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jarre.app.widgets;

import jarre.keyboard.IKeyboardPainter;
import jarre.keyboard.KeyData;
import jarre.keyboard.KeyboardPainterManager;
import jarre.midi.event.NoteMessageEvent;
import jarre.midi.event.NoteOffEvent;
import jarre.midi.event.NoteOnEvent;

import java.util.Arrays;
import java.util.HashMap;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Listener;

public class PianoKeyboard extends PianoAbstract {
    private Image background;
    private Color gray;
    private Color red;
    private Color white;
    private Color darkWhite;
    private Color black;
    private int pianoHeight;
    private boolean needUpdate = false;
    private Boolean needUpdateLock = new Boolean(true);
    private HashMap<Integer, NoteMessageEvent> noteEventMap = new HashMap<Integer, NoteMessageEvent>();
    private HashMap<Integer, RawNote> rawNoteMap = new HashMap<Integer, RawNote>();
    private int[] weight = new int[128];
    public PianoKeyboard(Composite parent, int style) {
	super(parent, style | SWT.NO_BACKGROUND | SWT.NO_FOCUS | SWT.NO_MERGE_PAINTS | SWT.DOUBLE_BUFFERED);
	gray = new Color(getDisplay(), 100, 100, 100);
	red = new Color(getDisplay(), 200, 0, 0);
	white = new Color(getDisplay(), 255, 255, 255);
	darkWhite = new Color(getDisplay(), 210, 210, 210);
	black = new Color(getDisplay(), 0, 0, 0);
	addListener(SWT.Resize, new Listener() {
	    @Override public void handleEvent(Event event) {
		Point size = computeSize(getSize().x, getSize().y, false);
		pianoHeight = size.y;
		createBackground();
	    }
	});
    }
    @Override public void dispose() {
	gray.dispose();
	red.dispose();
	white.dispose();
	darkWhite.dispose();
	black.dispose();
	if (background != null) background.dispose();
	super.dispose();
    }
    @Override public Point computeSize(int wHint, int hHint, boolean changed) {
	return KeyboardPainterManager.getCurrentPainter().computeSize(wHint, hHint);
    }
    public void setNoteOn(NoteOnEvent noteOnEvent) {
	setNeedUpdate(true);
	synchronized (noteEventMap) {
	    noteEventMap.put(calculateHash(noteOnEvent), noteOnEvent);
	}
    }
    public void setNoteOff(NoteOffEvent noteOffEvent) {
	setNeedUpdate(true);
	synchronized (noteEventMap) {
	    noteEventMap.remove(calculateHash(noteOffEvent));
	}
    }
    public boolean isRawNoteOn(int note) {
	return isRawNoteOn(note, -1);
    }
    public boolean isRawNoteOn(int note, int trackNum) {
	synchronized (rawNoteMap) {
	    return rawNoteMap.containsKey(calculateHash(note, trackNum));
	}
    }
    public void setRawNoteOn(int note) {
	setRawNoteOn(note, -1);
    }
    public void setRawNoteOn(int note, int trackNum) {
	setNeedUpdate(true);
	synchronized (rawNoteMap) {
	    rawNoteMap.put(calculateHash(note, trackNum), new RawNote(note, trackNum));
	}
    }
    public void setRawNoteOff(int note) {
	setRawNoteOff(note, -1);
    }
    public void setRawNoteOff(int note, int trackNum) {
	setNeedUpdate(true);
	synchronized (rawNoteMap) {
	    rawNoteMap.remove(calculateHash(note, trackNum));
	}
    }
    public void setAllNotesOff() {
	synchronized (noteEventMap) {
	    noteEventMap.clear();
	}
	redrawAndUpdate();
    }
    public void checkAndUpdate() {
	if (! isDisposed())
	    if (needUpdate()) {
		setNeedUpdate(false);
		redraw();
		update();
	    }
    }
    public void redrawAndUpdate() {
	if (! isDisposed()) {
	    setNeedUpdate(false);
	    redraw();
	    update();
	}
    }
    public boolean needUpdate() {
	synchronized (needUpdateLock) {
	    return needUpdate;
	}
    }
    public void setNeedUpdate(boolean needUpdate) {
	synchronized (needUpdateLock) {
	    this.needUpdate = needUpdate;
	}
    }
    protected void paintControl(PaintEvent e) {
	setNeedUpdate(false);
	GC gc = e.gc;
	int leftNote = KeyboardPainterManager.getLeftNote();
	int rightNote = KeyboardPainterManager.getRightNote();
	if (background != null) gc.drawImage(background, 0, 0);
	IKeyboardPainter kPainter = KeyboardPainterManager.getCurrentPainter();
	KeyData[] keyData = kPainter.getKeyData();
	Arrays.fill(weight, -1);
	synchronized (noteEventMap) {
	    for (NoteMessageEvent event: noteEventMap.values()) {		
		if (event.isVisible() && event.getNote() >= leftNote && event.getNote() <= rightNote) {
		    KeyData key = keyData[event.getNote()];
		    if (weight[event.getNote()] < event.getTick()) {
			weight[event.getNote()] = event.getTick();
			drawKey(gc, key, getTrackColor(event.getTrackNum()));
			if (event.getNote() == 60) drawBlock(gc, key);
		    }
		}
	    }
	}
	synchronized (rawNoteMap) {
	    for (RawNote event: rawNoteMap.values()) {
		if (event.note >= leftNote && event.note <= rightNote) {
		    KeyData key = keyData[event.note];
		    drawKey(gc, key, getTrackColor(event.trackNum));
		    if (event.note == 60) drawBlock(gc, key);
		}
	    }
	}
    }
    private int calculateHash(NoteMessageEvent noteMessageEvent) {
	return ((noteMessageEvent.getNote() & 0xFFFF) << 16) + (noteMessageEvent.getTrackNum() & 0xFFFF);
    }
    private int calculateHash(int note, int trackNum) {
	return ((note & 0xFFFF) << 16) + (trackNum & 0xFFFF);
    }
    private void drawBlock(GC gc, KeyData keyData) {
	int whiteKeyWidth = keyData.displayX3 - keyData.displayX;
	int blockWidth = (int)(whiteKeyWidth * 0.5);
	gc.setBackground(darkWhite);
	gc.fillRectangle(keyData.displayX + (whiteKeyWidth - blockWidth) / 2, pianoHeight - blockWidth * 2, blockWidth, blockWidth);
    }
    private void createBackground() {
	Point size = getSize();
	int leftNote = KeyboardPainterManager.getLeftNote();
	int rightNote = KeyboardPainterManager.getRightNote();
	IKeyboardPainter kPainter = KeyboardPainterManager.getCurrentPainter();
	KeyData[] keyData = kPainter.getKeyData();
	if (background != null) background.dispose();
	background = new Image(getDisplay(), size.x, size.y);
	GC gc = new GC(background);
	gc.setBackground(gray);
	gc.fillRectangle(0, 0, size.x, size.y);
	gc.setBackground(red);
	gc.fillRectangle(0, 0, KeyboardPainterManager.getWidth(), getRedHeight());
	for (int i = leftNote; i <= rightNote; i++) {
	    KeyData key = keyData[i];
	    if (key.type.isWhiteKey) {
		gc.setClipping(0, getRedHeight(), KeyboardPainterManager.getWidth(), KeyboardPainterManager.getHeight());
		kPainter.drawWhiteKey(gc, white, key);
		gc.setClipping((Rectangle)null);
	    }
	    else
		kPainter.drawBlackKey(gc, black, key);
	    if (i == 60) drawBlock(gc, key);
	}
	KeyData key = keyData[rightNote];
	gc.setBackground(black);
	int whiteKeyWidth = key.displayX3 - key.displayX;
	gc.fillRectangle(key.displayX + whiteKeyWidth, 0, getSize().x - KeyboardPainterManager.getWidth(), getSize().y);
	gc.dispose();
    }
    private void drawKey(GC gc, KeyData keyData, RGB rgb) {
	Color color = new Color(this.getDisplay(), rgb);
	IKeyboardPainter kPainter = KeyboardPainterManager.getCurrentPainter();
	if (keyData.type.isWhiteKey) {
	    gc.setClipping(0, getRedHeight(), KeyboardPainterManager.getWidth(), KeyboardPainterManager.getHeight());
	    kPainter.drawWhiteKey(gc, color, keyData);
	    gc.setClipping((Rectangle)null);
	}
	else kPainter.drawBlackKey(gc, color, keyData);
	color.dispose();
    }
    private int getRedHeight() {
	int h = (int)(pianoHeight * 0.02);
	if (h < 2) return 2;
	else return h;
    }
    private class RawNote {
	public final int note;
	public final int trackNum;
	public RawNote(int note, int trackNum) {
	    this.note = note;
	    this.trackNum = trackNum;
	}
    }
}
