/*  Programmer:	Piotr Borek
 *  E-mail: piotrborek@op.pl
 *  Copyright 2009 Piotr Borek
 *
 *  Distributed under the terms of the GPL (GNU Public License)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jarre.app.widgets;

import static jarre.i18n.I18n._;
import jarre.keyboard.IKeyboardPainter;
import jarre.keyboard.KeyData;
import jarre.keyboard.KeyType;
import jarre.keyboard.KeyboardPainterManager;
import jarre.midi.MySequence;
import jarre.midi.MyTrack;
import jarre.midi.event.EventMessage;
import jarre.midi.event.NoteOffEvent;
import jarre.midi.event.NoteOnEvent;
import jarre.midi.sequencer.PianoSequencer;
import jarre.midi.sequencer.SeqDrawEvent;
import jarre.midi.sequencer.SeqNoteEvent;
import jarre.midi.sequencer.SeqTrackEvent;
import jarre.widgets.windows.AppWindow;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.FontData;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Listener;

public class PianoTape extends PianoAbstract implements SeqNoteEvent, SeqTrackEvent, SeqDrawEvent {
    private int speed = 100;
    private Font font;
    private Image background;
    private Point size;
    private Color blue;
    private Color darkGray;
    private Color gray;
    private PianoSequencer pianoSequencer;
    private int displayQuarterLength = 64;
    private float tapeScale;
    private List<TrackDrawer> trackList = new ArrayList<TrackDrawer>();
    private boolean canRedraw = true;
    private Boolean canRedrawLock = new Boolean(true);
    private int lastCurrentPos;
    public PianoTape(Composite parent, int style) {
	super(parent, style | SWT.NO_BACKGROUND | SWT.NO_FOCUS | SWT.NO_MERGE_PAINTS | SWT.DOUBLE_BUFFERED);
	blue = new Color(getDisplay(), 190, 201, 221);
	darkGray = new Color(getDisplay(), 90, 90, 90);
	gray = new Color(getDisplay(), 160, 160, 160);
	FontData fontData = getDisplay().getSystemFont().getFontData()[0];
	font = new Font(getDisplay(), fontData.getName(), 14, SWT.BOLD);
	addListener(SWT.Resize, new Listener() {
	    @Override public void handleEvent(Event event) {
		size = getSize();
		createBackground();
	    }
	});
    }
    public void setSequencer(PianoSequencer pianoSequencer) {
	this.pianoSequencer = pianoSequencer;
	this.pianoSequencer.addNoteEvent(this);
	this.pianoSequencer.addDrawEvent(this);
	this.pianoSequencer.addTrackEvent(this);
	runSequenceChanged(pianoSequencer.getSequence());
    }
    public void setTapeScale(final int percent) {
	getDisplay().syncExec(new Runnable() {
	    @Override public void run() {
		displayQuarterLength = (64 * percent) / 100;
		if (displayQuarterLength < 16) displayQuarterLength = 16;
		if (displayQuarterLength > 1024) displayQuarterLength = 1024;
		tapeScaleChanged();
		createBackground();
		redraw();
		update();
	    }
	});
    }
    public int getTapeScale() {
	return (100 * displayQuarterLength) / 64;
    }
    @Override public void dispose() {
	blue.dispose();
	darkGray.dispose();
	gray.dispose();
	if (background != null) background.dispose();
	font.dispose();
	super.dispose();
    }
    public void redrawAndUpdate() {
	setCanRedraw(false);
	if (! isDisposed()) {
	    redraw();
	    update();
	}
    }
    public void setCanRedraw(boolean canRedraw) {
	synchronized (canRedrawLock) {
	    this.canRedraw = canRedraw;
	}
    }
    public boolean canRedraw() {
	synchronized (canRedrawLock) {
	    return canRedraw;
	}
    }
    public void setSpeed(int speed) {
	this.speed = speed;
	if (canRedraw()) redrawAndUpdate();
    }
    @Override public void endOfTracksEvent() {

    }
    @Override public void allNotesOffEvent() {
    }
    @Override public void noteOffEvent(NoteOffEvent noteOff) {
    }
    @Override public void noteOnEvent(NoteOnEvent noteOn) {
	trackList.get(noteOn.getTrackNum()).noteOnEvent(noteOn);
    }
    @Override public void sequenceChanged(MySequence sequence) {
	runSequenceChanged(sequence);
    }
    @Override public void drawSequence(int currentPos) {
	if (canRedraw()) {
	    currentPos = (int)(currentPos / tapeScale);
	    if (lastCurrentPos != currentPos) {
		lastCurrentPos = currentPos;
		setCanRedraw(false);
		redrawAndUpdate();
	    }
	}
    }
    @Override public void moveSequence(int currentPos) {
	while (!canRedraw()) AppWindow.readAndDispatch(getShell());
	setCanRedraw(false);
	redrawAndUpdate();
    }
    @Override protected void paintControl(PaintEvent event) {
	GC gc = event.gc;
	if (background != null) gc.drawImage(background, 0, 0);
	for (TrackDrawer drawer: trackList) drawer.paintControl(event);
	gc.setFont(font);
	gc.setForeground(getDisplay().getSystemColor(SWT.COLOR_WHITE));
	gc.drawString(_("Speed: ") + speed + " %", 10, 10, true);
	setCanRedraw(true);
    }
    private void runSequenceChanged(MySequence sequence) {
	if (sequence == null) return;
	trackList.clear();
	MyTrack[] tracks = sequence.getMyTracks();
	for (int i = 0; i < tracks.length; i++) trackList.add(new TrackDrawer(tracks[i], i));
	tapeScaleChanged();
	redrawAndUpdate();
    }
    private void tapeScaleChanged() {
	MySequence sequence = pianoSequencer.getSequence();
	if (sequence != null)
	    tapeScale = (float)sequence.getResolution() / displayQuarterLength;
    }
    private void createBackground() {
	int leftNote = KeyboardPainterManager.getLeftNote();
	int rigthNote = KeyboardPainterManager.getRightNote();
	IKeyboardPainter kPainter = KeyboardPainterManager.getCurrentPainter();
	KeyData[] keyData = kPainter.getKeyData();
	if (background != null) background.dispose();
	background = new Image(getDisplay(), size.x, size.y);
	GC gc = new GC(background);
	gc.setBackground(gray);
	gc.fillRectangle(0, 0, KeyboardPainterManager.getWidth(), size.y);
	for (int i = leftNote; i <= rigthNote; i++) {
	    KeyData key = keyData[i];
	    if (key.type == KeyType.C) {
		gc.setBackground(blue);
		gc.fillRectangle(key.displayX1, 0, key.displayX2 - key.displayX1, size.y);
	    }
	    if (! key.type.isWhiteKey) {
		gc.setBackground(darkGray);
		gc.fillRectangle(key.displayX1, 0, key.displayX2 - key.displayX1, size.y);
	    }
	}
	gc.setForeground(blue);
	int k = 0;
	int quarterLength = (displayQuarterLength * pianoSequencer.getNumerator()) / 4;
	for (int i = size.y - 1; i > 0; i = i - quarterLength, k++) {
	    if ((k % pianoSequencer.getNumerator()) == 0) gc.setForeground(getDisplay().getSystemColor(SWT.COLOR_BLACK));
	    else gc.setForeground(blue);
	    gc.drawLine(0, i, KeyboardPainterManager.getWidth() - 1, i);
	}
	KeyData key = keyData[rigthNote];
	gc.setBackground(getDisplay().getSystemColor(SWT.COLOR_BLACK));
	gc.fillRectangle(key.displayX3, 0, getSize().x - KeyboardPainterManager.getWidth(), getSize().y);
	gc.dispose();
    }
    private class TrackDrawer {
	private MyTrack track;
	private int trackNum;
	private LinkedList<NoteOnEvent> messageQueue = new LinkedList<NoteOnEvent>();
	public TrackDrawer(MyTrack track, int trackNum) {
	    this.track = track;
	    this.trackNum = trackNum;
	}
	public void noteOnEvent(NoteOnEvent noteOn) {
	    synchronized (messageQueue) {
		messageQueue.add(noteOn);
	    }
	}
	public void paintControl(PaintEvent event) {
	    MySequence sequence = pianoSequencer.getSequence();
	    if (! sequence.getTrackVisible(trackNum)) return;
	    GC gc = event.gc;
	    Color color = new Color(gc.getDevice(), getTrackColor(trackNum));
	    gc.setBackground(color);
	    gc.setForeground(getDisplay().getSystemColor(SWT.COLOR_BLACK));
	    int currentPos = pianoSequencer.getCurrentPos();
	    synchronized (messageQueue) {
		for (Iterator<NoteOnEvent> iter = messageQueue.iterator(); iter.hasNext(); ) {
		    NoteOnEvent noteOn = iter.next();
		    if (noteOn.getTick() > currentPos) {
			iter.remove();
			continue;
		    }
		    int length = noteOn.getTick() + noteOn.getLength() - currentPos;
		    if (length > 0) drawKeyNote(gc, noteOn.getNote(), 0, (int)(length / tapeScale));
		    else iter.remove();
		}
	    }
	    int height = (int)(size.y * tapeScale);
	    for (int index = 0; index < height; index++) {
		EventMessage[] events = track.getMessages(currentPos + index);
		for (EventMessage message: events) {
		    switch (message.type) {
			case NOTE_ON:
			    int i = (int)(index / tapeScale);
			    NoteOnEvent noteOn = message.getEvent();
			    drawKeyNote(gc, noteOn.getNote(), i, (int)(noteOn.getLength() / tapeScale));
			    break;
		    }
		}
	    }
	}
	private void drawKeyNote(GC gc, int note, int ly, int length) {
	    int y1 = size.y - ly - 1;
	    int y2 = y1 - length;
	    IKeyboardPainter kPainter = KeyboardPainterManager.getCurrentPainter();
	    KeyData key = kPainter.getKeyData()[note];
	    if (key == null) return;
	    gc.fillRectangle(key.displayX1, y1, key.displayX2 - key.displayX1 - 2, y2 - y1 + 1);
	    gc.drawRectangle(key.displayX1, y1, key.displayX2 - key.displayX1 - 2, y2 - y1);
	}
    }
}
