/*  Programmer:	Piotr Borek
 *  E-mail: piotrborek@op.pl
 *  Copyright 2009 Piotr Borek
 *
 *  Distributed under the terms of the GPL (GNU Public License)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jarre.keyboard;

import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Point;

public class SimpleKeyboardPainter implements IKeyboardPainter {
    private KeyData[] keyData = new KeyData[0];
    private int leftNote;
    private int rightNote;
    private int width;
    private int height;
    private float whiteKeyWidth;
    private float blackKeyWidth;
    @Override public void setKeyRange(int leftNote, int rightNote) {
	this.leftNote = leftNote;
	this.rightNote = rightNote;
	initializeKeyData();
    }
    @Override public void setKeyboardSize(int width, int height) {
	this.width = width;
	this.height = height;
	initializeKeyData();
    }
    @Override public Point computeSize(int wHint, int hHint) {
	int keyCount = KeyboardPainterManager.countWhiteKeys(leftNote, rightNote);
	if (wHint == SWT.DEFAULT && hHint == SWT.DEFAULT) {
	    wHint = 13 * keyCount;
	    hHint = 84;
	}
	else if (wHint == SWT.DEFAULT) {
	    int keyWidth = (int)(hHint / 6.5);
	    wHint = keyWidth * keyCount;
	} else {
	    int keyWidth = wHint / keyCount;
	    hHint = (int)(keyWidth * 6.5);
	}
	return new Point(wHint, hHint);
    }
    @Override public KeyData[] getKeyData() {
	synchronized (keyData) {
	    return keyData;
	}
    }
    @Override public void drawBlackKey(GC gc, Color keyColor, KeyData keyData) {
	gc.setBackground(keyColor);
	gc.fillRectangle(keyData.displayX1, 0, keyData.displayX2 - keyData.displayX1 - 1 , getBlackKeyHeight());
    }
    @Override public void drawWhiteKey(GC gc, Color keyColor, KeyData keyData) {
	gc.setBackground(keyColor);
	gc.fillRectangle(keyData.displayX1, 0, keyData.displayX2 - keyData.displayX1 - 1, height);
	gc.fillRectangle(keyData.displayX, getBlackKeyHeight() + 1, keyData.displayX3 - keyData.displayX - 1, height - getBlackKeyHeight() - 1);
    }
    private float getWhiteKeyWidth() {
	return (float)width / KeyboardPainterManager.countWhiteKeys(leftNote, rightNote);
    }
    private float getBlackKeyWidth() {
	return (getWhiteKeyWidth() * 3) / 5;
    }

    private int getBlackKeyHeight() {
	return (int)(height * 0.6);
    }
    private KeyData newKeyData(Vector4f v, KeyType keyType, int note) {
	return new KeyData((int)v.a, (int)v.b, (int)v.c, (int)v.d, keyType, note);
    }
    private void initializeKeyData() {
	KeyData[] newKeyData = new KeyData[128];
	whiteKeyWidth = getWhiteKeyWidth();
	blackKeyWidth = getBlackKeyWidth();
	int leftNote = this.leftNote;
	KeyType keyType = KeyboardPainterManager.getKeyType(leftNote);
	float lFix = 0;
	switch (keyType) {
	    case D:
	    case E:
	    case G:
	    case A:
	    case H:
		leftNote++;
		lFix = whiteKeyWidth;
	}
	Vector4f v = new Vector4f();
	v.set(-whiteKeyWidth, -blackKeyWidth, 0.0f, 0.0f);
	for (int i = 0; i < leftNote; i++) computeNextVector(v, i);
	float max = v.getMax();
	v.add(lFix - max);
	if (this.leftNote != leftNote)
	    newKeyData[this.leftNote] = new KeyData((int)v.a, (int)v.a, (int)v.c, (int)v.d, keyType, this.leftNote);
	for (int i = leftNote; i <= rightNote; i++) {
	    computeNextVector(v, i);
	    keyType = KeyboardPainterManager.getKeyType(i);
	    newKeyData[i] = newKeyData(v, keyType, i);
	    if (i == rightNote)
		newKeyData[i] = new KeyData(newKeyData[i].displayX, newKeyData[i].displayX1, width, width, keyType, i);
	}
	synchronized (keyData) {
	    keyData = newKeyData;
	}
    }
    private void computeNextVector(Vector4f v, int nextNote) {
	final float a = 5.0f / 16.0f;
	final float b = 1.0f - a;
	KeyType keyType = KeyboardPainterManager.getKeyType(nextNote);
	if (keyType.isWhiteKey) {
	    v.a = v.d;
	    v.d += whiteKeyWidth;
	    switch (keyType) {
	    	case C:
	    	case F:
	    	    v.b = v.a + blackKeyWidth * a;
		    v.c = v.d - blackKeyWidth * b;
	    	    break;
	    	case D:
	    	    v.b = v.a + blackKeyWidth * a;
		    v.c = v.d - blackKeyWidth * a;
		    break;
	    	case E:
	    	case H:
	    	    v.b = v.a + blackKeyWidth * b;
		    v.c = v.d - blackKeyWidth * a;
	    	    break;
	    	case G:
	    	    v.b = v.a + blackKeyWidth * a;
		    v.c = v.d - blackKeyWidth / 2;
		    break;
	    	case A:
	    	    v.b = v.a + blackKeyWidth / 2;
	    	    v.c = v.d - blackKeyWidth * a;
	    	    break;
	    }
	} else {
	    v.b = v.c;
	    v.c = v.c + blackKeyWidth;
	}
	switch (keyType) {
	    case C:
	    case F:
		v.b = v.a;
		break;
	    case E:
	    case H:
		v.c = v.d;
		break;
	}
    }
    private final class Vector4f  {
	public float a;
	public float b;
	public float c;
	public float d;
	public void set(float a, float b, float c, float d) {
	    this.a = a;
	    this.b = b;
	    this.c = c;
	    this.d = d;
	}
	public float getMax() {
	    float q = (a > b) ? a: b;
	    float w = (c > d) ? c: d;
	    return (q > w) ? q: w;
	}
	public void add(float x) {
	    a += x;
	    b += x;
	    c += x;
	    d += x;
	}
	@Override public String toString() {
	    return "Vector4 = [" + a + ", " + b + ", " + c + ", " + d +"]";
	}
    }
}
