/*
 *	Programmer:	Piotr Borek
 *	E-mail:		piotrborek@op.pl
 *	Copyright 2010  Piotr Borek
 *
 *	Distributed under the terms of the GPL (GNU Public License)
 *
 *   Jarre is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jarre.midi;

import static jarre.i18n.I18n._;
import jarre.Config;
import jarre.widgets.utils.Utils;

import java.util.LinkedList;
import java.util.List;

import javax.sound.midi.InvalidMidiDataException;
import javax.sound.midi.MidiDevice;
import javax.sound.midi.MidiMessage;
import javax.sound.midi.MidiUnavailableException;
import javax.sound.midi.Receiver;
import javax.sound.midi.ShortMessage;
import javax.sound.midi.Transmitter;

import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Shell;

public class MidiSystem {
    public static void searchForMidiDevices() {
	final Shell shell = new Shell();
	final Display display = shell.getDisplay();
	shell.setLayout(new GridLayout(2, false));
	shell.addListener(SWT.Close, new Listener() {
	    @Override public void handleEvent(Event event) {
		event.doit = false;
	    }
	});
	Label label = new Label(shell, SWT.NONE);
	label.setImage(display.getSystemImage(SWT.ICON_WORKING));
	label = new Label(shell, SWT.NONE);
	label.setText(_("Please wait. Searching for midi devices."));
	Thread thread = new Thread() {
	    @Override public void run() {
		MidiDriver.initialize();
		display.syncExec(new Runnable() {
		    @Override public void run() {
			shell.dispose();
		    }
		});
	    }
	};
	thread.start();
	shell.pack();
	shell.open();
	while (!shell.isDisposed()) {
	    if (!display.readAndDispatch()) display.sleep();
	}
    }
    public static void initialize() {
	boolean in_exception = false;
	boolean out_exception = false;
	closeMidi();
	String inName = Config.xmlConfig.getString("config.device.in", null);
	String outName = Config.xmlConfig.getString("config.device.out", null);
	String driverName = Config.xmlConfig.getString("config.device.driver", MidiDriver.MIDI_JAVA);
	if (inName != null) {
	    inMidi = MidiDriver.getInputByName(driverName, inName);
	    if (inMidi != null) {
		try {
		    inMidi.open();
		} catch (MidiUnavailableException e) {
		    in_exception = true;
		}
	    }
	}
	if (outName != null) {
	    outMidi = MidiDriver.getOutputByName(driverName, outName);
	    if (outMidi != null) {
		try {
		    outMidi.open();
		} catch (MidiUnavailableException e) {
		    out_exception = true;
		}
	    }
	}
	outReceiver = new ReceiverProxy();
	outReceiver.initialize();
	for (ReceiverProxy receiver: receiverList)
	    receiver.initialize();
	for (TransmitterProxy transmitter: transmitterList)
	    transmitter.initialize();
	if (Config.xmlConfig.getBoolean("config.device[@redirect]", false)) {
	    if (inMidi != null && outMidi != null && !in_exception && !out_exception) {
		getTransmitter().setReceiver(getReceiver());
	    }
	}
	if (in_exception || out_exception) {
	    String message = _("Error.\n");
	    if (in_exception) message += _("Input device unaviable.\n");
	    if (out_exception) message += _("Output device unaviable.\n");
	    Utils.showMessageDialog(_("Information"), message, SWT.ICON_INFORMATION | SWT.OK);
	}
    }
    public static void allSoundOff() {
	for (int i = 0; i < 16; i++) channelSoundOff(i);
    }
    public static void channelSoundOff(int channel) {
	try {
	    ShortMessage message = new ShortMessage();
	    message.setMessage(ShortMessage.CONTROL_CHANGE, channel, 0x7B /* All notes off */, 0);
	    dispatchMessageDirect(message, -1);
	    message = new ShortMessage();
	    message.setMessage(ShortMessage.CONTROL_CHANGE, channel, 0x40 /* Sustain off*/ , 0);
	    dispatchMessageDirect(message, -1);
	} catch (InvalidMidiDataException e) {
	    System.out.println("Never should be here. SequencerAbstract line 136");
	    e.printStackTrace();
	}
    }
    public static void dispatchMessageDirect(MidiMessage message, int time) {
	outReceiver.send(message, time);
    }
    public static void closeMidi() {
	if (inMidi != null)
	    if (inMidi.isOpen()) inMidi.close();
	if (outMidi != null)
	    if (outMidi.isOpen()) outMidi.close();
    }
    public static Transmitter getTransmitter() {
	TransmitterProxy proxy = new TransmitterProxy();
	proxy.initialize();
	transmitterList.add(proxy);
	return proxy;
    }
    public static Receiver getReceiver() {
	ReceiverProxy proxy = new ReceiverProxy();
	proxy.initialize();
	receiverList.add(proxy);
	return proxy;
    }
    private static List<ReceiverProxy> receiverList = new LinkedList<ReceiverProxy>();
    private static List<TransmitterProxy> transmitterList = new LinkedList<TransmitterProxy>();
    private static MidiDevice inMidi = null;
    private static MidiDevice outMidi = null;
    private static ReceiverProxy outReceiver;
    private static class TransmitterProxy implements Transmitter {
	Receiver receiver;
	Transmitter transmitter;
	public void initialize() {
	    if (inMidi == null) return;
	    if (transmitter != null) transmitter.close();
	    try {
		transmitter = inMidi.getTransmitter();
		if (receiver != null)
		    setReceiver();
	    } catch (MidiUnavailableException e) {
	    }
	}
	@Override public void close() {
	    transmitterList.remove(this);
	    if (transmitter == null) return;
	    transmitter.close();
	}
	@Override public Receiver getReceiver() {
	    return receiver;
	}
	@Override public void setReceiver(Receiver receiver) {
	    this.receiver = receiver;
	    if (transmitter == null) return;
	    setReceiver();
	}
	private void setReceiver() {
	    transmitter.setReceiver(receiver);
	}
    }
    private static class ReceiverProxy implements Receiver {
	Receiver receiver;
	public void initialize() {
	    if (outMidi == null) return;
	    if (receiver != null) receiver.close();
	    try {
		receiver = outMidi.getReceiver();
	    } catch (MidiUnavailableException e) {
	    }
	}
	@Override public void close() {
	    receiverList.remove(this);
	    if (receiver == null) return;
	    receiver.close();
	}
	@Override public void send(MidiMessage message, long timeStamp) {
	    if (receiver == null) return;
	    receiver.send(message, timeStamp);
	}	
    }
}
