/*  Programmer:	Piotr Borek
 *  E-mail: piotrborek@op.pl
 *  Copyright 2009 Piotr Borek
 *
 *  Distributed under the terms of the GPL (GNU Public License)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jarre.midi;

import jarre.midi.event.EventEnum;
import jarre.midi.event.EventMessage;
import jarre.midi.event.NoteOffEvent;
import jarre.midi.event.NoteOnEvent;
import jarre.midi.event.ProgramChangeEvent;
import jarre.midi.event.TempoChangeEvent;
import jarre.midi.event.TimeSignatureEvent;

import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import javax.sound.midi.MetaMessage;
import javax.sound.midi.MidiEvent;
import javax.sound.midi.MidiMessage;
import javax.sound.midi.ShortMessage;
import javax.sound.midi.SysexMessage;
import javax.sound.midi.Track;

public class MyTrack {
    public final int length;
    private static final EventMessage[] nullMessage = new EventMessage[] { new EventMessage(EventEnum.NONE) };
    private static final EventMessage[] exceptionMessage = new EventMessage[] { new EventMessage(EventEnum.ARRAY_INDEX_OUT_OF_BOUNDS) };
    private HashMap<Integer, EventMessage[]> messages = new HashMap<Integer, EventMessage[]>();
    private LinkedList<EventMessage> messageQueue = new LinkedList<EventMessage>();
    private List<TempoChangeEvent> tempoList = new LinkedList<TempoChangeEvent>();
    private int currentTick;
    private HashMap<Integer, NoteOnEvent> noteOnMap = new HashMap<Integer, NoteOnEvent>();
    private ShortMessage previousProgramChange = null;
    private MetaMessage previousTempoChange = null;
    private MetaMessage previousTimeSignature = null;
    private boolean trackVisible = true;
    private boolean playBackground = false;
    private final int trackNum;
    private final MySequence sequence;
    public MyTrack(MySequence sequence, Track track, int trackNum) {
	this.trackNum = trackNum;
	this.sequence = sequence;
	int size = track.size() - 1;
	int lastTick = (int)track.get(size).getTick();
	length = lastTick + 1;
	parseTrack(track);
    }
    public EventMessage[] getMessages(int timeTick) {
	EventMessage[] array;
	try {
	    array = messages.get(timeTick);
	} catch (ArrayIndexOutOfBoundsException e) {
	    return exceptionMessage;
	}
	if (array == null) return nullMessage;
	else return array;
    }
    public int getTrackNum() {
	return trackNum;
    }
    public List<TempoChangeEvent> getTempoList() {
	return Collections.unmodifiableList(tempoList);
    }
    public boolean getTrackVisible() {
        return trackVisible;
    }
    public void setTrackVisible(boolean trackVisible) {
        this.trackVisible = trackVisible;
    }
    public boolean getPlayBackground() {
        return playBackground;
    }
    public void setPlayBackground(boolean playBackground) {
        this.playBackground = playBackground;
    }
    private void parseTrack(Track track) {
	MidiEvent event;
	MidiMessage message;
	long timeTick;
	int length = track.size();
	for (int i = 0; i < length; i++) {
	    event = track.get(i);
	    message = event.getMessage();
	    timeTick = event.getTick();
	    while (currentTick < timeTick) {
		copyMessages(currentTick);
		currentTick++;
	    }
	    if (message instanceof MetaMessage) messageQueue.add(parseMetaMessage((MetaMessage)message));
	    else if (message instanceof SysexMessage) messageQueue.add(parseSysexMessage((SysexMessage)message));
	    else if (message instanceof ShortMessage) {
		EventMessage eventMessage = parseShortMessage((ShortMessage)message);
		if (eventMessage != null) messageQueue.add(eventMessage);
	    }
	}
	copyMessages(currentTick);
    }
    private void copyMessages(int index) {
	if (index >= 0 && ! messageQueue.isEmpty()) {
	    messages.put(index, messageQueue.toArray(new EventMessage[0]));
	    messageQueue.clear();
	}
    }
    private EventMessage parseMetaMessage(MetaMessage message) {
	switch (message.getType()) {
	    case 0x51:
		TempoChangeEvent tempoChangeEvent = new TempoChangeEvent(sequence, message, previousTempoChange, trackNum, currentTick);
		previousTempoChange = message;
		tempoList.add(tempoChangeEvent);
		return new EventMessage(tempoChangeEvent);
	    case 0x58:
		TimeSignatureEvent timeSignatureEvent = new TimeSignatureEvent(sequence, message, previousTimeSignature, trackNum, currentTick);
		previousTimeSignature = message;
		return new EventMessage(timeSignatureEvent);
	    case 0x2F:
		return new EventMessage(EventEnum.END_OF_TRACK);
	    default:
		return new EventMessage(message);
	}
    }
    private EventMessage parseSysexMessage(SysexMessage message) {
	return new EventMessage(message);
    }
    private EventMessage parseShortMessage(ShortMessage message) {
	switch (message.getCommand()) {
	    case ShortMessage.NOTE_ON:
		int velocity = message.getData2();
		if (velocity != 0) {
		    NoteOnEvent noteOnEvent = new NoteOnEvent(sequence, message, trackNum, currentTick, 0);
		    noteOnMap.put(noteOnEvent.getNote(), noteOnEvent);
		    return new EventMessage(noteOnEvent);
		}
	    case ShortMessage.NOTE_OFF:
		NoteOnEvent noteOnEvent = noteOnMap.remove(message.getData1());
		if (noteOnEvent == null) return null;
		noteOnEvent.setLength(currentTick - noteOnEvent.getTick());
		return new EventMessage(new NoteOffEvent(sequence, message, noteOnEvent, trackNum, currentTick));
	    case ShortMessage.PROGRAM_CHANGE:
		ProgramChangeEvent programChangeEvent = new ProgramChangeEvent(sequence, message, previousProgramChange, trackNum, currentTick);
		previousProgramChange = message;
		return new EventMessage(programChangeEvent);
	    default:
		return new EventMessage(message);
	}
    }
}
