/*  Programmer:	Piotr Borek
 *  E-mail: piotrborek@op.pl
 *  Copyright 2010 Piotr Borek
 *
 *  Distributed under the terms of the GPL (GNU Public License)
 *
 *  Jarre program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jarre.midi.alsa;

import java.io.File;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import javax.sound.midi.InvalidMidiDataException;
import javax.sound.midi.MidiDevice;
import javax.sound.midi.MidiMessage;
import javax.sound.midi.MidiUnavailableException;
import javax.sound.midi.Receiver;
import javax.sound.midi.ShortMessage;
import javax.sound.midi.Transmitter;

public class AlsaMidiDevice implements MidiDevice {
    private static final String LIBNAME = "jarre-alsajni-";
    public static List<MidiDevice> getDevices() {
	return deviceList;
    }
    public static void initialize() {
	String arch = System.getProperty("sun.arch.data.model");
	if (arch == null) System.getProperty("com.ibm.vm.bitmode");
	if (arch.equals("32")) arch = "i686";
	else if (arch.equals("64")) arch = "x86_64";
	File file = new File("native/lib" + LIBNAME + arch + ".so");
	if (file.exists())
	    System.loadLibrary(LIBNAME + arch);
	else
	    return;
	long handle = AlsaJNI.open_client("Jarre", false);
	long info = AlsaJNI.new_info();
	while (AlsaJNI.query_next_client(handle, info) >= 0) {
	    long port_info = AlsaJNI.new_port_info(info);
	    while (AlsaJNI.query_next_port(handle, port_info) >= 0) {
		MidiDevice device = new AlsaMidiDevice(info, port_info);
		deviceList.add(device);
	    }
	    AlsaJNI.free_port_info(port_info);
	}
	AlsaJNI.free_info(info);
	AlsaJNI.close_client(handle);
    }
    AlsaMidiDevice(long info, long port_info) {
	String name = AlsaJNI.port_get_name(port_info);
	String vendor = "Alsa";
	String description = AlsaJNI.info_get_name(info);
	String version = "1.0";
	this.info = new Info(name, vendor, description, version) {
	};
	readable = AlsaJNI.port_can_read(port_info);
	writable = AlsaJNI.port_can_write(port_info);
	addr = AlsaJNI.info_get_id(info);
	port = AlsaJNI.port_get_id(port_info);	
    }
    @Override public void close() {
	isOpened = false;
	handleLock.lock();
	try {
	    if (handle == 0) return;
	    if (writable) AlsaJNI.disconnect_port_to(handle, 0, addr, port);
	    if (readable) AlsaJNI.disconnect_port_from(handle, 0, addr, port);
	    if (handleCounter == 0) return;
	    handleCounter--;
	    if (handleCounter == 0) {
		AlsaJNI.close_client(handle);
		handle = 0;
	    }
	} finally {
	    handleLock.unlock();
	}
    }
    @Override public Info getDeviceInfo() {
	return info;
    }
    @Override public int getMaxReceivers() {
	if (writable) return -1;
	else return 0;
    }
    @Override public int getMaxTransmitters() {
	if (readable) return -1;
	else return 0;
    }
    @Override public long getMicrosecondPosition() {
	return 0;
    }
    @Override public Receiver getReceiver() throws MidiUnavailableException {
	Receiver receiver = new Receiver() {
	    @Override public void close() {
		receiverList.remove(this);
	    }
	    @Override public void send(MidiMessage msg, long time) {
		if (msg instanceof ShortMessage) {
		    ShortMessage shortMessage = (ShortMessage)msg;
		    handleLock.lock();
		    try {
			if (handle == 0) return;
			switch (shortMessage.getCommand()) {
		    	    case ShortMessage.NOTE_ON:
		    		AlsaJNI.send_note_on(handle, addr, port, shortMessage.getChannel(), shortMessage.getData1(), shortMessage.getData2());
		    		break;
		    	    case ShortMessage.NOTE_OFF:
		    		AlsaJNI.send_note_off(handle, addr, port, shortMessage.getChannel(), shortMessage.getData1(), shortMessage.getData2());
		    		break;
		    	    case ShortMessage.PROGRAM_CHANGE:
		    		AlsaJNI.send_program_change(handle, addr, port, shortMessage.getChannel(), shortMessage.getData1());
		    		break;
		    	    case ShortMessage.CONTROL_CHANGE:
		    		AlsaJNI.send_controller(handle, addr, port, shortMessage.getChannel(), shortMessage.getData1(), shortMessage.getData2());
		    		break;
		    	    case ShortMessage.PITCH_BEND:
		    		AlsaJNI.send_pitchbend(handle, addr, port, shortMessage.getChannel(), shortMessage.getData1());
		    		break;
		    	    case ShortMessage.CHANNEL_PRESSURE:
		    		AlsaJNI.send_chanpress(handle, addr, port, shortMessage.getChannel(), shortMessage.getData1());
		    		break;
			}
		    } finally {
			handleLock.unlock();
		    }
		}
	    }	    
	};
	receiverList.add(receiver);
	return receiver;
    }
    @Override public List<Receiver> getReceivers() {
	return receiverList;
    }
    @Override public Transmitter getTransmitter() throws MidiUnavailableException {
	Transmitter transmitter = new Transmitter() {
	    private Receiver receiver;
	    @Override public void close() {
		synchronized (transmitterList) {
		    transmitterList.remove(this);
		}
	    }
	    @Override synchronized public Receiver getReceiver() {
		return receiver;
	    }
	    @Override synchronized public void setReceiver(Receiver receiver) {
		this.receiver = receiver;
	    }
	};
	synchronized (transmitterList) {
	    transmitterList.add(transmitter);
	}
	return transmitter;
    }
    @Override public List<Transmitter> getTransmitters() {
	return transmitterList;
    }
    @Override public boolean isOpen() {
	return isOpened;
    }
    @Override public void open() throws MidiUnavailableException {
	handleLock.lock();
	try {
	    if (handleCounter == 0) handle = AlsaJNI.open_client("Jarre", true);
	    if (handle == 0) return;
	    handleCounter++;
	    isOpened = true;
	    if (writable) AlsaJNI.connect_port_to(handle, 0, addr, port);
	    if (readable) {
		AlsaJNI.connect_port_from(handle, 0, addr, port);
		startReadThread();
	    }
	} finally {
	    handleLock.unlock();
	}
    }
    private void startReadThread() {
	readThread = new Thread() {
	    @Override public void run() {
		while (isOpened) {
		    handleLock.lock();
		    try {
			if (handle == 0) return;
			if (AlsaJNI.event_pending(handle) > 0) {
			    long event = AlsaJNI.retrieve_event(handle);
			    int type = AlsaJNI.event_get_type(event);
			    int channel = AlsaJNI.event_get_channel(event);
			    int data1 = AlsaJNI.event_get_data1(event);
			    int data2 = AlsaJNI.event_get_data2(event);
			    AlsaJNI.free_event(event);
			    ShortMessage message = new ShortMessage();
			    int command = -1;
			    switch (type) {
			    	case AlsaEventType.NOTEON:
			    	    command = ShortMessage.NOTE_ON;
			    	    break;
			    	case AlsaEventType.NOTEOFF:
			    	    command = ShortMessage.NOTE_OFF;
			    	    break;
			    	case AlsaEventType.CONTROLLER:
			    	    command = ShortMessage.CONTROL_CHANGE;
			    	    break;
			    	case AlsaEventType.PGMCHANGE:
			    	    command = ShortMessage.PROGRAM_CHANGE;
			    	    break;
			    	case AlsaEventType.CHANPRESS:
			    	    command = ShortMessage.CHANNEL_PRESSURE;
			    	    break;
			    	case AlsaEventType.PITCHBEND:
			    	    command = ShortMessage.PITCH_BEND;
			    	    break;
			    	case AlsaEventType.SONGPOS:
			    	    command = ShortMessage.SONG_POSITION_POINTER;
			    	    break;
			    	case AlsaEventType.SONGSEL:
			    	    command = ShortMessage.SONG_SELECT;
			    	    break;
			    	case AlsaEventType.TIMESIGN:
			    	    command = ShortMessage.TIMING_CLOCK;
			    	    break;
			    }
			    if (command < 0) continue;
			    try {
				message.setMessage(command, channel, data1, data2);
				synchronized (transmitterList) {
				    for (Transmitter t: transmitterList)
					t.getReceiver().send(message, 0);
				}
			    } catch (InvalidMidiDataException e) {
			    }
			}
		    } finally {
			handleLock.unlock();
		    }
		}
	    }
	};
	readThread.start();	
    }
    private Thread readThread;
    private int addr;
    private int port;
    private volatile boolean isOpened = false;
    private Info info;
    private boolean readable;
    private boolean writable;
    private LinkedList<Transmitter> transmitterList = new LinkedList<Transmitter>();
    private LinkedList<Receiver> receiverList = new LinkedList<Receiver>();
    //
    private static LinkedList<MidiDevice> deviceList = new LinkedList<MidiDevice>();
    private static long handle;
    private static Lock handleLock = new ReentrantLock();
    private static int handleCounter = 0;
}
