/*  Programmer:	Piotr Borek
 *  E-mail: piotrborek@op.pl
 *  Copyright 2009 Piotr Borek
 *
 *  Distributed under the terms of the GPL (GNU Public License)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jarre.midi.sequencer;

import jarre.midi.MySequence;
import jarre.midi.MyTrack;
import jarre.midi.event.EventEnum;
import jarre.midi.event.EventMessage;
import jarre.midi.event.NoteMessageEvent;
import jarre.midi.event.NoteOffEvent;
import jarre.midi.event.NoteOnEvent;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

public class SequencerHelper {
    private final List<TrackHelper> trackList = new ArrayList<TrackHelper>();
    private int tick = 0;
    protected SequencerHelper(MySequence sequence) {
	for (MyTrack track: sequence.getMyTracks())
	    trackList.add(new TrackHelper(track));
    }
    public synchronized void setCurrentTick(int tick) {
	this.tick = tick;
    }
    public synchronized int getCurrentTick() {
	return tick;
    }
    public synchronized List<EventMessage> getCurrentMessages() {
	boolean eof = true;
	LinkedList<EventMessage> list = new LinkedList<EventMessage>();
	for (TrackHelper helper: trackList)
	    eof &= helper.processCurrentMessages(list);
	if (eof) return null;
	else return list;
    }
    public synchronized List<EventMessage> getActiveMessages() {
	LinkedList<EventMessage> list = new LinkedList<EventMessage>();
	for (TrackHelper helper: trackList)
	    helper.processActiveMessages(list);
	return list;
    }
    public synchronized List<EventMessage> goForward(int ticks) {
	LinkedList<EventMessage> list = new LinkedList<EventMessage>();
	for (TrackHelper helper: trackList) {
	    helper.goForward(ticks);
	    helper.processActiveMessages(list);
	}
	this.tick += ticks;
	return list;
    }
    public synchronized List<EventMessage> goBackward(int ticks) {
	LinkedList<EventMessage> list = new LinkedList<EventMessage>();
	for (TrackHelper helper: trackList) {
	    helper.goBackward(ticks);
	    helper.processActiveMessages(list);
	}
	this.tick -= ticks;
	if (this.tick < 0) this.tick = 0;
	return list;
    }
    private class TrackHelper {
	private final HashMap<Integer, EventMessage> trackNoteOnMap = new HashMap<Integer, EventMessage>();
	private EventMessage tempoChangeMessage;
	private EventMessage programChangeMessage;
	private EventMessage timeSignatureMessage;
	public final MyTrack track;
	protected TrackHelper(MyTrack track) {
	    this.track = track;
	}
	public boolean processCurrentMessages(List<EventMessage> list) {
	    if (! track.getTrackVisible() && ! track.getPlayBackground()) return false;
	    for (EventMessage eventMessage: track.getMessages(tick))
		switch (eventMessage.type) {
		    case NONE:
			break;
		    case ARRAY_INDEX_OUT_OF_BOUNDS:
		    case END_OF_TRACK:
			return true;
		    case NOTE_ON:
			putNote(eventMessage);
			list.add(eventMessage);
			break;
		    case NOTE_OFF:
			removeNote(eventMessage);
			list.add(eventMessage);
			break;
		    default:
			list.add(eventMessage);
		}
	    return false;
	}
	public void processActiveMessages(List<EventMessage> list) {
	    if (! track.getTrackVisible() && ! track.getPlayBackground()) return;
	    if (tempoChangeMessage != null) list.add(tempoChangeMessage);
	    if (programChangeMessage != null) list.add(programChangeMessage);
	    if (timeSignatureMessage != null) list.add(timeSignatureMessage);
	    for (EventMessage eventMessage: trackNoteOnMap.values())
		list.add(eventMessage);
	}
	public void goForward(int ticks) {
	    tempoChangeMessage = null;
	    programChangeMessage = null;
	    timeSignatureMessage = null;
	    int end = tick + ticks;
	    for (int i = tick; i <= end; i++) {
		EventMessage[] events = track.getMessages(i);
		for (EventMessage message: events) {
		    switch (message.type) {
			case NOTE_ON:
			    putNote(message);
			    break;
			case NOTE_OFF:
			    removeNote(message);
			    break;
			default:
			    processMessages(message);
		    }
		}
	    }
	}
	public void goBackward(int ticks) {
	    tempoChangeMessage = null;
	    programChangeMessage = null;
	    timeSignatureMessage = null;
	    int start = tick - ticks;
	    if (start > 0) {
		for (int i = tick; i > start; i--) {
		    EventMessage[] events = track.getMessages(i);
		    for (EventMessage message: events)
			if (message.type == EventEnum.NOTE_ON)
			    removeNote(message);
		    for (EventMessage message: events) {
			switch (message.type) {
			    case NOTE_OFF:
				NoteOffEvent noteOff = message.getEvent();
				trackNoteOnMap.put(noteOff.getNote(), new EventMessage(noteOff.getNoteOnEvent()));
				break;
			    default:
				processMessages(message);
			}

		    }
		}
	    } else trackNoteOnMap.clear();
	}
	private void processMessages(EventMessage message) {
	    switch (message.type) {
		case TEMPO_CHANGE:
		    tempoChangeMessage = message;
		    break;
		case TIME_SIGNATURE:
		    timeSignatureMessage = message;
		    break;
		case PROGRAM_CHANGE:
		    programChangeMessage = message;
		    break;
	    }
	}
	private void putNote(EventMessage message) {
	    NoteOnEvent noteOn = message.getEvent();
	    trackNoteOnMap.put(noteOn.getNote(), message);
	}
	private void removeNote(EventMessage message) {
	    NoteMessageEvent noteEvent = message.getEvent();
	    trackNoteOnMap.remove(noteEvent.getNote());
	}
    }
}
